import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.text.*;

/**
 * @author <A HREF="mailto:weyer@kagi.com">Steve Weyer</A>
 * Copyright (c) 2001, S. Weyer. All right reserved.
 * @version Personal Java 1.0
 * <A HREF="http://www.kagi.com/weyer/">Steve's other apps</A>
 *
 */


public class Mines
  extends Frame
  implements
    Runnable,
    ItemListener,     // for Choice
    ActionListener    // for Button
  {

  public static final boolean DEBUG = false;

  private static final String[] NUM_MINES = {"10", "20", "30", "40", "50"};

  private Label timeLabel, numLabel;
  private Button newButton;
  private Choice numList;
  private MineCanvas canvas;
  private Thread timer;

  int time = 0, mines = 10;

  public static void main(String[] args) {
    new Mines();  // ??? params
  }

  public Mines() {
    super("Mines");

    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent ev) { System.exit(0); } });

    setLayout(new BorderLayout());
    Panel p = new Panel();

    newButton = new Button("New");
    newButton.addActionListener(this);
    p.add(newButton);

    timeLabel = new Label("00:00", Label.CENTER);
    p.add(timeLabel);

    numList = new Choice();
    for (int i=0; i < NUM_MINES.length; i++)
      numList.addItem(NUM_MINES[i]);
    //numList.select(0);
    numList.addItemListener(this);
    p.add(numList);

    numLabel = new Label("    ", Label.CENTER);
    p.add(numLabel);
    add(p, BorderLayout.NORTH);

    canvas = new MineCanvas(this);
    canvas.setSize(192,192);  // initially
    add(canvas, BorderLayout.CENTER);
    //pack();
    Dimension screenSize = getToolkit().getScreenSize();
    setSize(Math.min(screenSize.width,192), Math.min(screenSize.height,240));
    setVisible(true);

    timer = new Thread(this);
    timer.start();
  }

  public void newGame(int n) { // called at startup, New, <#mines>
    updateTime(0);
    updateMines(mines = n);
    canvas.newGame(mines);
  }

  public void updateTime(int newTime) {
    time = newTime;
    int mins = time / 60, secs = time % 60;
    timeLabel.setText(String.valueOf(mins) + ':' + ((secs < 10) ? "0" : "") + secs);
  }

  public void updateMines(int newMine) {
    numLabel.setText(String.valueOf(newMine));
  }

  public void endGame(String msg) { // called after win/lose
    getToolkit().beep();
    numLabel.setText(msg);
  }

  public void run() { // via timer & Runnable
    while (true) {  // ??? stop condition other than exit
      try {
        switch (canvas.gameState) {
          case MineCanvas.OK:
            updateTime(time+1);
            break;
          case MineCanvas.INIT: // very 1st time (avoid async img refresh problems)???
            newGame(mines);
          default:
        };
        timer.sleep(1000);
      }
      catch (InterruptedException ex) {
        System.err.println("thread: " + ex);
      };
    }
  }

  public void actionPerformed(ActionEvent ev) {
    //if (ev.getSource() == newButton)
      newGame(mines);
  }

  public void itemStateChanged(ItemEvent ev) {
    //if (ev.getSource() == numList)
      newGame(Integer.parseInt(numList.getSelectedItem()));
  }
}


class MineCanvas
  extends Canvas
  implements MouseListener
  {
  // overall game state
  public final static int INIT = -2;
  public final static int START = -1;
  public final static int OK = 0;
  public final static int WON = 2;
  public final static int LOST = 3;

  final static int GRID_SIZE = 16; // w,h of Images
  Image upSquare, downSquare, bomb, xedBomb, flag, boom;
  Mines mines;
  Graphics drawg = null;
  int cols, rows, mines_left;
  int gameState = INIT;
  int lastX, lastY;

  // board state
  static final byte UNTOUCHED = 0;
  static final byte MARKED = 1;
  static final byte UNCOVERED = 2;

  protected boolean[][] grid;
  public byte[][] boardState;
  public byte[][] neighbors;
  private Random rand;

  public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
    switch (infoflags) {
      case ALLBITS:
        //System.out.println("loaded: " + img);
        // success. do nothing
        imgsToLoad--;
        return false;
      case ERROR:
      case ABORT:
        //System.err.println("not loaded: " + img);
        return false;
    };
    return true;  // SOMEBITS, FRAMEBITS, WIDTH, HEIGHT
  }

  public int imgsToLoad;

  public Image getImage(String name) {
    java.net.URL imgURL = getClass().getResource(name);
    Image img = null;
    if (imgURL != null)
      prepareImage(img = getToolkit().getImage(imgURL), this);  // createImage slightly better but 1.2+
      // imageUpdate is called async via imageUpdate
    //else System.err.println("failed to load Image: " + name);
    return img;
  }

  public MineCanvas(Mines app) {
    imgsToLoad = 6;
    upSquare = getImage("upsqr16.gif");
    downSquare = getImage("insqr16.gif");
    bomb =  getImage("bomb16.gif");
    flag = getImage("flag16.gif");
    xedBomb = getImage("xedbomb16.gif");
    boom = getImage("boom16.gif");

    mines = app;
    rand = new Random();
    addMouseListener(this);
  }

  public int nextInt(int n) { // >= 1.2 rand.nextInt(10));
    return (int) (rand.nextFloat() * n);
  }

  public void newGame(int n) {
    if (imgsToLoad > 0)
      return;

    gameState = START;
    // get each time since window may have changed size
    Dimension d = getSize();
    cols = d.width  / GRID_SIZE;
    rows = d.height / GRID_SIZE;
//System.out.println("cols,rows: " + cols + ' ' + rows);

    grid = new boolean[cols][rows];     //false
    boardState = new byte[cols][rows];  //UNTOUCHED
    neighbors = new byte[cols][rows];   //0

    mines_left = n; // ??? reduce if too few squares
    int x, y;
    while (n > 0) {
      x = nextInt(cols);
      y = nextInt(rows);
      if (grid[x][y])
	continue;
//System.out.println("bomb: " + n + ' ' + x + ' ' + y);
      grid[x][y] = true;
      n--;
    };

    int xmin, xmax, ymin = 0, ymax;
    for (y = 0; y < rows; y++) {
      ymax = ((y == rows-1) ? y : y+1);
      xmin = 0;
      for (x = 0; x < cols; x++) {
	if (grid[x][y]) {
          xmax = ((x == cols-1) ? x : x+1);
	  for (int i = xmin; i <= xmax; i++)
	    for (int j = ymin; j <= ymax; j++)
	      neighbors[i][j] += 1;
	};
        xmin = x;
      };
      ymin = y;
    };

  repaint();
  }


  public void paint(Graphics g) {
    if (grid == null)   // no new game
      return;
    drawg = g.create();  // init here for later local updates

    for (int y = 0; y < rows; y++)
      for (int x = 0; x < cols; x++)
        squareChanged(x, y);
  }

  public void squareChanged (int x, int y) {    // mark, uncover, or just repainting
    boolean isMine = grid[x][y];
    Image img = null;
    String txt = null;
    switch (boardState[x][y]) {
      case UNCOVERED:
        if (isMine)
          img = (x == lastX && y == lastY) ? boom : bomb;
        else {
          img = downSquare;
          int nbors = neighbors[x][y];
          if (nbors != 0)
            txt = String.valueOf(nbors);
        };
        break;
      case UNTOUCHED:
        img = (gameState == LOST && isMine) ? bomb : upSquare;
        break;
      case MARKED:
        img = (gameState == LOST && !isMine) ? xedBomb : flag;
        //break;
      default:
    };

    int gx = x*GRID_SIZE, gy = y*GRID_SIZE;
    drawg.drawImage(img, gx, gy, this);
    if (txt != null) {
      drawg.setColor(Color.black);
      drawg.drawString(txt, gx+3, gy+13);  // offset: left+3, -bottom-3
      drawg.setColor(Color.white);
    };
  }

  public void mouseEntered(MouseEvent ev) {
    //ignore
  }
  public void mouseExited(MouseEvent ev) {
    //ignore
  }
  public void mouseClicked(MouseEvent ev) {
    // handled via press/release
  }

  public void mousePressed(MouseEvent ev) {
    if (gameState == START)
      gameState = OK; // starts updating time
    int curX = ev.getX() / GRID_SIZE;
    int curY = ev.getY() / GRID_SIZE;

    if (gameState != OK || curX < 0 || curX >= cols || curY < 0 || curY >= rows)
      return;

    // save for mouseReleased
    lastX = curX;
    lastY = curY;
  }

  public void mouseReleased(MouseEvent ev) {
    int curX = ev.getX() / GRID_SIZE;
    int curY = ev.getY() / GRID_SIZE;
    if (gameState != OK || curX < 0 || curX >= cols || curY < 0 || curY >= rows)
      return;

    byte st = boardState[lastX][lastY];
    if (curX == lastX && curY == lastY) {
      // same square. uncover it (if not already marked/uncovered)
      if (st == UNTOUCHED) {
        uncover(lastX, lastY);
        if (grid[lastX][lastY]) {
          gameState = LOST;
          mines.endGame("Lost!");
          repaint();
          return;
        };
      };
    }
    else {
      // drag to diff square. (un)mark the orig square and decrement(inc) the count
      switch (st) {
        case UNTOUCHED:
          st = MARKED;    // flag
          mines_left--;
          break;
        case MARKED:
          st = UNTOUCHED; // unflag
          mines_left++;
          break;
        default:
          return;
        };
      boardState[lastX][lastY] = st;
      squareChanged(lastX, lastY);
      mines.updateMines(mines_left);
    };

    if (Won()) {
      gameState = WON;
      mines.endGame("Won!");
    };
  }

  public void uncover(int x, int y) {
    // only recurse on squares that haven't been touched before.
    //  LU | U | RU
    // ------------
    //  L  | - | R
    // ------------
    //  LD | D | RD
    if (boardState[x][y] == UNTOUCHED) {
      boardState[x][y] = UNCOVERED;
      squareChanged(x,y);
      if (neighbors[x][y] == 0) {
        int x1 = x+1, y1 = y+1;
        if (x > 0) {
          uncover(x-1,y); // left
          if (y > 0)
            uncover(x-1,y-1); // left-up
          if (y1 < rows)
            uncover(x-1,y1);  // left-down
        };
        if (x1 < cols) {
          uncover(x1,y); // right
          if (y > 0)
            uncover(x1,y-1);  // right-up
          if (y1 < rows)
            uncover(x1,y1);  // right-down
        };
        if (y > 0)
          uncover(x,y-1);  // up
        if (y1 < rows)
          uncover(x,y1);  // down
      };
    };
  }

  public boolean Won() {
    byte st;
    for (int x = 0; x < cols; x++)
      for (int y = 0; y < rows; y++)
        if ((st = boardState[x][y]) == UNTOUCHED)
	  return false;
	else if (st == MARKED && ! grid[x][y])
	  return false;

    return true;
  }
}